'use strict';

const path = require('path');
const gulp = require('gulp');
const url = require('url');
const Handlebars = require('handlebars');
const jsdom = require('jsdom');
const glob = require('glob');
const read = require('read-file');
const reload = require('reload');
const jsonfile = require('jsonfile');
const fs = require('fs');
const copydir = require('copy-dir');
const request = require('request-json');

module.exports.Helper = class Helper {
    constructor(config) {
        this.watchable = [];
        this.workingDir = config.workingDir || '.';
        this.components = config.components || {};
        this.routes = config.routes || {};
        this.layouts = config.layouts || {};
        this.regex = /\$(\w+)/g; // api url variable replacement
    }

    getLayout(name, reload) {
        if (this.layouts.hasOwnProperty(name)) {
            const layout = this.layouts[name];
            if (reload) {
                layout.template = read.sync(layout.file, 'utf8');
            }
            return layout
        }
    }

    getComponent(name, reload) {
        if (this.components.hasOwnProperty(name)) {
            const component = this.components[name];
            if (reload) {
                this.createComponent(component.file)
            }
            return this.components[name]
        }
    }

    createComponent(filePath) {
        const component = {
            file: filePath,
            name: path.basename(filePath, '.html'),
            template: '',
        };

        const template = read.sync(component.file, 'utf8');
        const dom = new jsdom.JSDOM(template, {
            includeNodeLocations: true
        });
        const body = dom.window.document.head;
        for (let i = 0; i < body.children.length; i++) {
            let child = body.children[i];
            if (child.nodeName === "TEMPLATE") {
                const loc = dom.nodeLocation(child);
                const start = loc.startTag.endOffset;
                const end = loc.endOffset - start - 11;
                component.template = template.substr(start, end);
            } else if (child.nodeName === "SCRIPT") {
                component.script = child.innerHTML
            }
        }

        Handlebars.registerPartial(component.name, component.template);
        component.precompiled = Handlebars.precompile(component.template);
        this.components[component.name] = component;
    }

    createLayout(filePath) {
        const layout = {
            file: filePath,
            name: path.basename(filePath, '.html'),
        };
        layout.template = read.sync(layout.file, 'utf8');
        this.layouts[layout.name] = layout;
    }

    parseLayout(pattern) {
        this.watchable.push({pattern: pattern, type: 'layout'});
        const files = glob.sync(pattern);
        for (let i = 0; i < files.length; i++) {
            this.createLayout(files[i])
        }
    }

    parse(pattern) {
        this.watchable.push({pattern: pattern, type: 'component'});
        const files = glob.sync(pattern);
        for (let i = 0; i < files.length; i++) {
            this.createComponent(files[i])
        }
    }

    buildRoute(h, route, basePath, parents) {
        this.routeCount += 1;
        route.parents = parents;
        route.id = this.routeCount;
        let newPath = path.join(basePath, route.path);
        if (route.path === '/') {
            newPath += '/'
        }
        if (!this.hasOwnProperty(newPath)) {
            h[newPath] = [];
        }
        if (parents) {
            h[newPath] = h[newPath].concat(parents)
        }
        if (!route.children || route.children.length === 0) {
            h[newPath].push(route)
        }
        if (route.children && route.children.length > 0) {
            let ps = parents || [];
            ps.push(route);
            for (let i = 0; i < route.children.length; i++) {
                h = this.buildRoute(h, route.children[i], newPath, ps)
            }
        }
        return h
    }

    changeDetected(watchable) {
        return change => {
            if (watchable.type === 'component') {
                this.createComponent(change.path)
            } else if (watchable.type === 'layout') {
                this.createLayout(change.path)
            }
            if (this.reload) {
                Handlebars.registerPartial('handlebars', this.packagedHandlebars());
                Handlebars.registerPartial('customElements', this.packagedCustomElements());
                this.reload.reload()
            }
        }
    }

    setupExpress(app, opt) {
        this.routeCount = -1;
        this.handle = {}; // map of array of routes
        for (let i = 0; i < this.routes.length; i++) {
            this.handle = this.buildRoute(this.handle, this.routes[i], "/", [])
        }

        for (const p in this.handle) {
            if (this.handle.hasOwnProperty(p)) {
                const rs = this.handle[p];
                app.get(p, this.handleRoute(p, rs))
            }
        }

        if (opt) {
            if (opt.hasOwnProperty('liveReload') && opt.liveReload === true) {
                this.reload = reload(app);
                for (let i = 0; i < this.watchable.length; i++) {
                    const w = this.watchable[i];
                    gulp.watch(w.pattern, this.changeDetected(w))
                }
            }
        }

        app.use('/components.js', (req, res, next) => {
            res.send(this.packagedHandlebars() + ";" + this.packagedCustomElements());
            next()
        });
    }

    handleRoute(p, rs) {
        return (req, res) => {
            return new Promise((resolve, reject) => {
                let context = {};
                let layoutName = 'index';

                if (rs && rs.length > 0) {
                    const r = rs[0];
                    if (r && r.hasOwnProperty('layout') && r.layout.length > 0) {
                        layoutName = r.layout
                    }
                }

                const layout = this.getLayout(layoutName);
                const registeredPages = [];
                let resolvedApiUri = '';

                let outletName = 'router-outlet';
                for (let i = 0; i < rs.length; i++) {
                    const r = rs[i];
                    if (r.hasOwnProperty('outlet') && r.outlet && r.outlet.length > 0) {
                        outletName = r.outlet
                    }
                    registeredPages.push(outletName);
                    Handlebars.registerPartial(outletName, this.components[r.component].template);
                    if (r.hasOwnProperty('api')) {
                        resolvedApiUri = r.api;
                        resolvedApiUri = resolvedApiUri.replace(this.regex, function (match, param) {
                            return match.replace('$' + param, req.params[param]);
                        });

                        let split = resolvedApiUri.split('?');
                        resolvedApiUri = split[0];
                        let resolvedQuery = '';
                        if (split.length > 1) {
                            resolvedQuery = split[1]
                        }

                        for (let k in req.query) {
                            if (req.query.hasOwnProperty(k)) {
                                resolvedQuery += '&' + req.query[k]
                            }
                        }

                        resolvedApiUri += '?' + resolvedQuery
                    }
                    if (r.hasOwnProperty('page')) {
                        context = Object.assign(context, r.page);
                    }
                }
                const fn = Handlebars.compile(layout.template);

                // get api data
                if (resolvedApiUri.length > 0) {
                    console.log('resolving ' + resolvedApiUri);
                    const client = request.createClient(resolvedApiUri);
                    client.get('', function (err, response, body) {
                        context.data = body;
                        res.send(fn(context));
                        // remove pages from cache
                        for (let i = 0; i < registeredPages.length; i++) {
                            Handlebars.unregisterPartial(registeredPages[i])
                        }
                        resolve()
                    });
                } else {
                    res.send(fn(context));
                    // remove pages from cache
                    for (let i = 0; i < registeredPages.length; i++) {
                        Handlebars.unregisterPartial(registeredPages[i])
                    }
                    resolve()
                }
            })
        }
    }

    packagedHandlebars() {
        const cArr = Object.keys(this.components).map(i => this.components[i]);
        return `(function(){'use strict';var template=Handlebars.template,templates=Handlebars.templates=Handlebars.templates || {};${cArr.map((component) => {
            return `templates['${component.name}'] = template(${component.precompiled});`
        }).join('')}})()`
    }

    packagedCustomElements() {
        const cArr = Object.keys(this.components).map(i => this.components[i]);
        return `(function(){'use strict';var cc=new function(){this.define=function(name,module){if(module&&module.hasOwnProperty('exports')){module.exports.prototype.template=function(c,o){return Handlebars.templates[name](c,o)};window.customElements.define(name,module.exports)}}};window['customComponents']=cc;${cArr.map((component) => {
            return component.script ? `cc.define('${component.name}',(function(){var module={};${component.script};return module})());` : ''
        }).join('')}})();`
    }

    build(dist) {
        mkDirByPathSync(dist);
        mkDirByPathSync(dist + '/components');
        mkDirByPathSync(dist + '/public');
        // write custom.js
        fs.writeFileSync(dist + "/public/components.js", this.packagedHandlebars() + ";" + this.packagedCustomElements());
        // copy static files
        copydir.sync(this.workingDir + '/public', dist + '/public');
        // build manifest.json file
        const manifest = {
            'imports': [],
            'routes': this.routes,
        };
        for (let name in this.components) {
            if (this.components.hasOwnProperty(name)) {
                const newFilePath = 'components/' + name + '.component.html';

                // write file
                fs.writeFileSync(dist + '/' + newFilePath, this.components[name].template);

                manifest.imports.push({
                    'path': newFilePath,
                    'name': name,
                })
            }
        }

        for (let name in this.layouts) {
            if (this.layouts.hasOwnProperty(name)) {
                const newFilePath = 'components/' + name + '.layout.html';

                // write file
                fs.writeFileSync(dist + '/' + newFilePath, this.layouts[name].template);

                manifest.imports.push({
                    'path': newFilePath,
                    'name': name,
                    'layout': true,
                })
            }
        }

        jsonfile.writeFileSync(dist + '/manifest.json', manifest)
    }
};

function mkDirByPathSync(targetDir, {isRelativeToScript = false} = {}) {
    const sep = path.sep;
    const initDir = path.isAbsolute(targetDir) ? sep : '';
    const baseDir = isRelativeToScript ? __dirname : '.';
    targetDir.split(sep).reduce((parentDir, childDir) => {
            const curDir = path.resolve(baseDir, parentDir, childDir);
            try {
                fs.mkdirSync(curDir);
            } catch (err) {
                if (err.code !== 'EEXIST') {
                    throw err;
                }
            }
            return curDir;
        },
        initDir
    )
}
